# This file is @generated by <https://github.com/liblaf/copier-python>.
# DO NOT EDIT!

# ref: <https://github.com/ofek/hatch-vcs>
__version__: str
__version_tuple__: tuple[int | str, ...]
version_tuple: tuple[int | str, ...]
version: str



from ._define import define, frozen
from ._field_specifiers import field, static
from ._filters import AuxData, combine, combine_leaves, partition, partition_leaves
from ._flatten import register_fieldz

__all__ = [
    "AuxData",
    "combine",
    "combine_leaves",
    "define",
    "field",
    "frozen",
    "partition",
    "partition_leaves",
    "register_fieldz",
    "static",
]



from collections.abc import Callable, Sequence

import fieldz
import jax.tree_util as jtu

from ._codegen import codegen


def register_fieldz[T: type](
    cls: T,
    data_fields: Sequence[str] | None = None,
    meta_fields: Sequence[str] | None = None,
) -> T:
    if data_fields is None:
        data_fields = _filter_fieldz(cls, static=False)
    if meta_fields is None:
        meta_fields = _filter_fieldz(cls, static=True)
    flatten: Callable
    flatten_with_keys: Callable
    unflatten: Callable
    flatten, flatten_with_keys, unflatten = codegen(cls, data_fields, meta_fields)
    jtu.register_pytree_with_keys(cls, flatten_with_keys, unflatten, flatten)
    return cls


def _filter_fieldz(cls: type, *, static: bool) -> list[str]:
    return [
        f.name for f in fieldz.fields(cls) if f.metadata.get("static", False) == static
    ]



from collections.abc import Iterable
from typing import Any

import jax

from jarp.utils import is_array

from ._define import frozen

type Leaf = Any
type PyTreeDef[T] = Any


@frozen(static=True)
class AuxData[T]:
    static_fields: tuple[Any, ...]
    treedef: PyTreeDef[T]


def combine[T](dynamic_leaves: Iterable[Any], aux: AuxData[T]) -> T:
    leaves: list[Any] = combine_leaves(dynamic_leaves, aux.static_fields)
    return jax.tree.unflatten(aux.treedef, leaves)


def combine_leaves(
    dynamic_leaves: Iterable[Any], static_leaves: Iterable[Any]
) -> list[Any]:
    return [
        static_leaf if dynamic_leaf is None else dynamic_leaf
        for dynamic_leaf, static_leaf in zip(dynamic_leaves, static_leaves, strict=True)
    ]


def partition[T](obj: T) -> tuple[list[Any], AuxData[T]]:
    leaves: list[Any]
    treedef: PyTreeDef[T]
    leaves, treedef = jax.tree.flatten(obj)
    dynamic_leaves: list[Any]
    static_leaves: list[Any]
    dynamic_leaves, static_leaves = partition_leaves(leaves)
    aux: AuxData = AuxData(tuple(static_leaves), treedef)
    return dynamic_leaves, aux


def partition_leaves(leaves: Iterable[Any]) -> tuple[list[Any], list[Any]]:
    dynamic_leaves: list[Any] = []
    static_leaves: list[Any] = []
    for leaf in leaves:
        if is_array(leaf):
            dynamic_leaves.append(leaf)
            static_leaves.append(None)
        else:
            dynamic_leaves.append(None)
            static_leaves.append(leaf)
    return dynamic_leaves, static_leaves



import ast
import linecache
import types
from collections.abc import Callable, Iterable
from typing import Any, NamedTuple, Protocol

import jax.tree_util as jtu

type KeyEntry = Any
type Leaf = Any

_PREFIX: str = "_jarp_codegen_"
_CLS: str = f"{_PREFIX}cls"
_OBJECT_NEW: str = f"{_PREFIX}object_new"
_OBJECT_SETATTR: str = f"{_PREFIX}object_setattr"


class FlattenFunction(Protocol):
    def __call__(self, obj: Any) -> tuple[tuple[Leaf, ...], tuple[Any, ...]]: ...


class FlattenWithKeysFunction(Protocol):
    def __call__(
        self, obj: Any
    ) -> tuple[tuple[tuple[KeyEntry, Leaf], ...], tuple[Any, ...]]: ...


class UnflattenFunction(Protocol):
    def __call__(self, aux: tuple[Any, ...], children: tuple[Leaf, ...]) -> Any: ...


class CodeGenResults(NamedTuple):
    tree_flatten: FlattenFunction
    tree_flatten_with_keys: FlattenWithKeysFunction
    tree_unflatten: UnflattenFunction


def codegen(
    cls: type, data_fields: Iterable[str], meta_fields: Iterable[str]
) -> CodeGenResults:
    return CodeGenResults(
        _codegen_flatten(cls, data_fields, meta_fields),
        _codegen_flatten_with_keys(cls, data_fields, meta_fields),
        _codegen_unflatten(cls, data_fields, meta_fields),
    )


def _add_dunder(cls: type, func: Callable) -> None:
    func.__module__ = cls.__module__


def _codegen_flatten(
    cls: type, data_fields: Iterable[str], meta_fields: Iterable[str]
) -> FlattenFunction:
    children: ast.Tuple = _codegen_attrgetter(data_fields)
    aux: ast.Tuple = _codegen_attrgetter(meta_fields)
    source: ast.FunctionDef = ast.FunctionDef(
        name="tree_flatten",
        args=ast.arguments(
            posonlyargs=[],
            args=[ast.arg("obj")],
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
        ),
        body=[ast.Return(ast.Tuple([children, aux], ast.Load()))],
        decorator_list=[],
        returns=None,
        type_comment=None,
        type_params=[],
    )
    filename: str = _make_filename(cls, "tree_flatten")
    namespace: dict[str, Any] = {}
    _compile(source, filename, namespace)
    func: FlattenFunction = namespace["tree_flatten"]
    _add_dunder(cls, func)
    return func


def _codegen_flatten_with_keys(
    cls: type, data_fields: Iterable[str], meta_fields: Iterable[str]
) -> FlattenWithKeysFunction:
    children: ast.Tuple = ast.Tuple(
        [
            ast.Tuple(
                [
                    ast.Name(f"{_PREFIX}{name}_key", ast.Load()),
                    ast.Attribute(ast.Name("obj", ast.Load()), name, ast.Load()),
                ],
                ast.Load(),
            )
            for name in data_fields
        ],
        ast.Load(),
    )
    aux: ast.Tuple = _codegen_attrgetter(meta_fields)
    source: ast.FunctionDef = ast.FunctionDef(
        name="tree_flatten_with_keys",
        args=ast.arguments(
            posonlyargs=[],
            args=[ast.arg("obj")],
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
        ),
        body=[ast.Return(ast.Tuple([children, aux], ast.Load()))],
        decorator_list=[],
        returns=None,
        type_comment=None,
        type_params=[],
    )
    filename: str = _make_filename(cls, "tree_flatten_with_keys")
    namespace: dict[str, Any] = {
        f"{_PREFIX}{name}_key": jtu.GetAttrKey(name) for name in data_fields
    }
    _compile(source, filename, namespace)
    func: FlattenWithKeysFunction = namespace["tree_flatten_with_keys"]
    _add_dunder(cls, func)
    return func


def _codegen_unflatten(
    cls: type, data_fields: Iterable[str], meta_fields: Iterable[str]
) -> UnflattenFunction:
    body: list[ast.stmt] = [
        # obj = _object_new(_cls)
        ast.Assign(
            [ast.Name("obj", ast.Store())],
            ast.Call(
                ast.Name(_OBJECT_NEW, ast.Load()), [ast.Name(_CLS, ast.Load())], []
            ),
        )
    ]
    if cls.__setattr__ is object.__setattr__:
        # obj.a, obj.b, ... = aux
        body.extend(_codegen_unpack("aux", meta_fields))
        # obj.a, obj.b, ... = children
        body.extend(_codegen_unpack("children", data_fields))
    else:
        # bypass cls.__setattr__
        # _a, _b, ... = aux
        # _object_setattr(obj, "a", _a)
        body.extend(_codegen_setattr("aux", meta_fields))
        # _a, _b, ... = children
        # _object_setattr(obj, "a", _a)
        body.extend(_codegen_setattr("children", data_fields))
    # return obj
    body.append(ast.Return(ast.Name("obj", ast.Load())))
    source: ast.FunctionDef = ast.FunctionDef(
        name="tree_unflatten",
        args=ast.arguments(
            posonlyargs=[],
            args=[ast.arg("aux"), ast.arg("children")],
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
        ),
        body=body,
        decorator_list=[],
        returns=None,
        type_comment=None,
        type_params=[],
    )
    filename: str = _make_filename(cls, "tree_unflatten")
    namespace: dict[str, Any] = {
        _CLS: cls,
        _OBJECT_NEW: object.__new__,
        _OBJECT_SETATTR: object.__setattr__,
    }
    _compile(source, filename, namespace)
    func: UnflattenFunction = namespace["tree_unflatten"]
    _add_dunder(cls, func)
    return func


def _codegen_attrgetter(fields: Iterable[str]) -> ast.Tuple:
    return ast.Tuple(
        [
            ast.Attribute(ast.Name("obj", ast.Load()), name, ast.Load())
            for name in fields
        ],
        ast.Load(),
    )


def _codegen_unpack(arg_name: str, fields: Iterable[str]) -> list[ast.stmt]:
    return [
        # obj.a, obj.b, ... = arg
        ast.Assign(
            [
                ast.Tuple(
                    [
                        ast.Attribute(ast.Name("obj", ast.Load()), name, ast.Store())
                        for name in fields
                    ],
                    ast.Store(),
                )
            ],
            ast.Name(arg_name, ast.Load()),
        )
    ]


def _codegen_setattr(arg_name: str, fields: Iterable[str]) -> list[ast.stmt]:
    # My benchmarks show that unpacking is slightly faster than indexing,
    # so we use unpacking even for setattr.
    body: list[ast.stmt] = [
        # _a, _b, ... = arg
        ast.Assign(
            [
                ast.Tuple(
                    [ast.Name(f"{_PREFIX}{name}", ast.Store()) for name in fields],
                    ast.Store(),
                )
            ],
            ast.Name(arg_name, ast.Load()),
        )
    ]
    body.extend(
        [
            # _object_setattr(obj, "name", _name)
            ast.Expr(
                ast.Call(
                    func=ast.Name(_OBJECT_SETATTR, ast.Load()),
                    args=[
                        ast.Name("obj", ast.Load()),
                        ast.Constant(name),
                        ast.Name(f"{_PREFIX}{name}", ast.Load()),
                    ],
                    keywords=[],
                )
            )
            for name in fields
        ]
    )
    return body


def _compile(source: ast.FunctionDef, filename: str, namespace: dict[str, Any]) -> None:
    module: ast.Module = ast.Module([source], type_ignores=[])
    module: ast.Module = ast.fix_missing_locations(module)
    code: types.CodeType = compile(module, filename, "exec")
    exec(code, namespace)  # noqa: S102
    _update_linecache(module, filename)


def _make_filename(cls: type, func_name: str) -> str:
    return f"<jarp generated {func_name} {cls.__module__}.{cls.__qualname__}>"


def _make_keys(fields: Iterable[str]) -> dict[str, KeyEntry]:
    return {f"{_PREFIX}{name}_key": jtu.GetAttrKey(name) for name in fields}


def _update_linecache(source: ast.Module, filename: str) -> None:
    code: str = ast.unparse(source)
    linecache.cache[filename] = (
        len(code),  # size
        None,  # mtime
        code.splitlines(keepends=True),  # lines
        filename,  # fullname
    )



from . import tree
from ._jit import jit
from ._version import __version__, __version_tuple__
from .tree import define, field, frozen, static

__all__ = [
    "__version__",
    "__version_tuple__",
    "define",
    "field",
    "frozen",
    "jit",
    "static",
    "tree",
]



import functools
import types
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import Any, Protocol, Self, TypedDict, Unpack, overload

import jax

from jarp import tree

type _Dynamic = Iterable[Any]
type _Static[T] = tree.AuxData[T]


class JitOptions(TypedDict, total=False):
    filter: bool
    inline: bool


@overload
def jit[**P, T](
    fun: Callable[P, T], **kwargs: Unpack[JitOptions]
) -> Callable[P, T]: ...
@overload
def jit[**P, T](
    **kwargs: Unpack[JitOptions],
) -> Callable[[Callable[P, T]], Callable[P, T]]: ...
def jit[**P, T](fun: Callable[P, T] | None = None, **kwargs) -> Callable:
    if fun is None:
        return functools.partial(jit, **kwargs)
    filter_: bool = kwargs.pop("filter", False)
    if not filter_:
        return jax.jit(fun, **kwargs)

    fun_dynamic: _Dynamic
    fun_static: _Static[Callable[P, T]]
    fun_dynamic, fun_static = tree.partition(fun)
    inner: _Inner[T] = _Inner(fun_static)
    inner_jit: _InnerProtocol[T] = jax.jit(inner, static_argnums=(2,), **kwargs)
    outer: _Outer[P, T] = _Outer(fun_dynamic=fun_dynamic, inner=inner_jit)
    functools.update_wrapper(outer, fun)
    return outer


class _InnerProtocol[T](Protocol):
    def __call__(
        self, inputs_dynamic: _Dynamic, fun_dynamic: _Dynamic, inputs_static: _Static
    ) -> tuple[_Dynamic, _Static[T]]: ...


@tree.frozen(static=True)
class _Inner[T](_InnerProtocol[T]):
    fun_static: _Static[Callable[..., T]]

    def __call__(
        self, inputs_dynamic: _Dynamic, fun_dynamic: _Dynamic, inputs_static: _Static
    ) -> tuple[_Dynamic, _Static[T]]:
        fun: Callable[..., T] = tree.combine(fun_dynamic, self.fun_static)
        args: Sequence[Any]
        kwargs: Mapping[str, Any]
        args, kwargs = tree.combine(inputs_dynamic, inputs_static)
        outputs: T = fun(*args, **kwargs)
        outputs_dynamic: _Dynamic
        outputs_static: _Static[T]
        outputs_dynamic, outputs_static = tree.partition(outputs)
        return outputs_dynamic, outputs_static


@tree.define(slots=False, static=False)
class _Outer[**P, T]:
    fun_dynamic: _Dynamic
    inner: _InnerProtocol[T]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        inputs_dynamic: _Dynamic
        inputs_static: _Static
        inputs_dynamic, inputs_static = tree.partition((args, kwargs))
        outputs_dynamic: _Dynamic
        outputs_static: _Static[T]
        outputs_dynamic, outputs_static = self.inner(
            inputs_dynamic, self.fun_dynamic, inputs_static
        )
        return tree.combine(outputs_dynamic, outputs_static)

    @overload
    def __get__(self, instance: None, owner: type | None = None) -> Self: ...
    @overload
    def __get__(
        self, instance: object, owner: type | None = None
    ) -> Callable[..., T]: ...
    def __get__(self, instance: Any, owner: type | None = None) -> Callable[..., T]:
        if instance is None:
            return self
        return types.MethodType(self, instance)



from typing import Any

import numpy as np
from jax import Array
from jax._src.literals import TypedNdArray

_ARRAY_TYPES: tuple[type, ...] = (Array, TypedNdArray, np.ndarray, np.generic)


def is_array(element: Any) -> bool:
    # ref: <https://github.com/patrick-kidger/equinox/blob/d1718838dcfff2b0adc4f7a795f72da7bdbec1aa/equinox/_filters.py#L44-L46>
    return isinstance(element, _ARRAY_TYPES)



from __future__ import annotations

import functools
from collections.abc import Callable
from typing import Any, TypedDict, Unpack, dataclass_transform, overload

import attrs
import jax.tree_util as jtu
from liblaf import grapes

from ._field_specifiers import field, static
from ._flatten import register_fieldz


class DefineOptions(TypedDict, total=False):
    these: dict[str, Any] | None
    repr: bool
    unsafe_hash: bool | None
    hash: bool | None
    init: bool
    slots: bool
    frozen: bool
    weakref_slot: bool
    str: bool
    auto_attribs: bool
    kw_only: bool
    cache_hash: bool
    auto_exc: bool
    eq: bool | None
    order: bool | None
    auto_detect: bool
    getstate_setstate: bool | None
    on_setattr: attrs._OnSetAttrArgType | None
    field_transformer: attrs._FieldTransformer | None
    match_args: bool

    static: bool


@overload
@dataclass_transform(field_specifiers=(attrs.field, field, static))
def define[T: type](cls: T, /, **kwargs: Unpack[DefineOptions]) -> T: ...
@overload
@dataclass_transform(field_specifiers=(attrs.field, field, static))
def define[T: type](**kwargs: Unpack[DefineOptions]) -> Callable[[T], T]: ...
def define[T: type](maybe_cls: T | None = None, **kwargs) -> T | Callable:
    _warnings_hide = True
    if maybe_cls is None:
        return functools.partial(define, **kwargs)
    static: bool = kwargs.pop("static", False)
    if static and not kwargs.get("frozen", False):
        grapes.warnings.warn(
            "Defining a static class that is not frozen may lead to unexpected behavior."
        )
    cls: T = grapes.attrs.define(maybe_cls, **kwargs)
    if static:
        jtu.register_static(cls)
    else:
        register_fieldz(cls)
    return cls


@overload
@dataclass_transform(frozen_default=True, field_specifiers=(attrs.field, field, static))
def frozen[T: type](cls: T, /, **kwargs: Unpack[DefineOptions]) -> T: ...
@overload
@dataclass_transform(frozen_default=True, field_specifiers=(attrs.field, field, static))
def frozen[T: type](**kwargs: Unpack[DefineOptions]) -> Callable[[T], T]: ...
def frozen[T: type](maybe_cls: T | None = None, **kwargs) -> T | Callable:
    _warnings_hide = True
    if maybe_cls is None:
        return functools.partial(frozen, **kwargs)
    kwargs.setdefault("frozen", True)
    return define(maybe_cls, **kwargs)



from collections.abc import Mapping
from typing import Any

import attrs
import tlz
from liblaf import grapes


@grapes.wraps(attrs.field)
def field(**kwargs) -> Any:
    static: bool = kwargs.pop("static", False)
    if static:
        metadata: Mapping = kwargs.get("metadata", {})
        metadata = tlz.assoc(metadata, "static", static)
        kwargs["metadata"] = metadata
    return attrs.field(**kwargs)


@grapes.wraps(attrs.field)
def static(**kwargs) -> Any:
    kwargs.setdefault("static", True)
    return field(**kwargs)
